# A part of the AegisBlade Python Client Library
# Copyright (C) 2019 Thornbury Organization, Bryan Thornbury
# This file may be used under the terms of the GNU Lesser General Public License, version 2.1.
# For more details see: https://www.gnu.org/licenses/lgpl-2.1.html

import json


def json_encode(post_data):
    def get_object_dict(obj):
        if (type(obj) == bytes):
            raise TypeError

        objDict = dict()
        for k, v in obj.__dict__.items():
            if not k.startswith("_x_"):
                objDict[k] = v

        return objDict

    return json.dumps(post_data, default=get_object_dict,
                      sort_keys=True, indent=4)


def json_decode(json_str, obj_type):
    obj = obj_type()
    obj.__dict__ = json.loads(json_str)
    return obj



# A part of the AegisBlade Python Client Library
# Copyright (C) 2019 Thornbury Organization, Bryan Thornbury
# This file may be used under the terms of the GNU Lesser General Public License, version 2.1.
# For more details see: https://www.gnu.org/licenses/lgpl-2.1.html

import base64
import sys


class ApplicationContextFile(object):
    """Represents a file required for execution of the current application.

    This class encapsulates all details required to determine if the file has
    changed from previous iterations, and to upload the file if needed.

    Attributes:
        fileHash (str): The sha256 hash of the file's contents.

        filePathRelativeToAppContext (str): The relative path of the file to CWD.

        filePath (str): Absolute path of the file on the source machine.

        fileByteCount (int): The number of bytes in the file.

        _x_fileContents (str): Base64 encoded string of the file's contents. 
            The strange '_x_' naming is a convention used to prevent 
            automatically serializing this value to json.
    """
    
    fileHash = None                      # type: str
    """`str`: The sha256 hash of the file's contents."""

    filePathRelativeToAppContext = None  # type: str
    """`str`: The relative path of the file to CWD."""

    filePath = None                      # type: str
    """`str`: Absolute path of the file on the source machine."""

    fileByteCount = None                 # type: int
    """`int`: The number of bytes in the file."""

    _x_fileContents = None               # type: str
    """`str`: Base64 encoded string of the file's contents."""

    def __init__(self, file_path, file_path_relative_to_app_context, file_hash, file_contents, file_byte_count):
        self.fileHash = file_hash
        self.filePathRelativeToAppContext = file_path_relative_to_app_context
        self.filePath = file_path
        self.fileByteCount = file_byte_count

        self._x_fileContents = base64.b64encode(file_contents)
        
        if sys.version_info[0] == 3:
            self._x_fileContents = self._x_fileContents.decode('ascii')


# A part of the AegisBlade Python Client Library
# Copyright (C) 2019 Thornbury Organization, Bryan Thornbury
# This file may be used under the terms of the GNU Lesser General Public License, version 2.1.
# For more details see: https://www.gnu.org/licenses/lgpl-2.1.html

def readFile(filePath):
    with open(filePath, 'r') as f:
        return f.read()


class HostDefinition(object):
    def __init__(self, jobConfig):
        self.driver = jobConfig.host_driver
        self.hostAffinity = jobConfig.host_affinity

        self.driverOptions = {
            'instanceType': jobConfig.host_driver_options['instanceType'],
            'region': jobConfig.host_driver_options['region'],
            'useSpotInstance': jobConfig.host_driver_options['useSpotInstance'],
            'maxSpotPrice': jobConfig.host_driver_options['maxSpotPrice']
        }


class CreateApplicationPayload(object):
    def __init__(self, applicationExecutionContext, jobConfig):
        '''
        :type applicationExecutionContext: ApplicationExecutionContext
        '''
        self.applicationExecutionContext = applicationExecutionContext
        self.hostDefinition = HostDefinition(jobConfig)
        self.capabilities = jobConfig.capabilities


class CreateJobPayload(object):
    def __init__(self, taskInfo, applicationId, jobType, clientVersion, jobConfig):
        self.jobType = jobType
        self.clientVersion = clientVersion
        self.memoryMb = jobConfig.memory
        self.applicationId = applicationId
        self.serializedJobEntrypoint = taskInfo.serializeForInjector()
        self.hostDefinition = HostDefinition(jobConfig)


class UploadFilePayload(object):
    def __init__(self, applicationContextFile=None, fileContents=None, applicationId=None):
        self.applicationContextFile = applicationContextFile
        self.applicationGuid = applicationId
        self.fileContents = fileContents

class CreateDataStorePayload(object):
    def __init__(self, dataStoreName, dataStorageDriver, dataStorageDriverOptions):
        self.dataStoreName = dataStoreName
        self.dataStorageDriver = dataStorageDriver
        self.driverOptions = dataStorageDriverOptions




# A part of the AegisBlade Python Client Library
# Copyright (C) 2019 Thornbury Organization, Bryan Thornbury
# This file may be used under the terms of the GNU Lesser General Public License, version 2.1.
# For more details see: https://www.gnu.org/licenses/lgpl-2.1.html

from typing import Optional, Dict, Any
import os

from aegisblade._internal.environment import Environment

try:
  basestring
except NameError:
  basestring = str

class JobConfig(object):
    """Configuration container for a job.

    The JobConfig is an optional object that can be used to
    specify configuration that will change the way a job is 
    executed, determine host placement, or configure values
    for the backing host driver.

    Attributes:
        libraries (Dict str=str): A dictionary of `package=local_path` of local
            packages used inside the project from outside the 
            project's directory tree. These libraries will be 
            uploaded as part of your application and put on the
            `PYTHONPATH` during its execution.

        memory (int): The amount of memory in MB required by the job. Used
            to determine how many jobs will be run per host. Specifying
            an amount greater than the instance type specified in 
            `JobConfig.host_driver_options` (or the default instance type), will not
            change the instance type to accomodate the greater amount
            of memory, but will force only one job to run per host.

        host_driver (str): The backing driver to use in launching the hosts
            for the job. ("ec2" only currently).
        
        host_driver_options (Dict str=Any): A dict of options for the host driver such
            as "instanceType", "region", and "useSpotInstance".

        host_affinity (str): A string label used to group jobs on a host(s). If 
            specified, the job will only run on hosts with the same host
            affinity label. For example, assigning "host1" to jobs 1-4 and
            "host2" to jobs 5-8 will force both sets of jobs to run on 
            different hosts regardless of other factors.
        
        capabilities (List `aegisblade.capabilities.Capability`): An array of capabilities required by this job. See
            `aegisblade.capabilities.Capability` for more details.
    """

    _default_memory = 1024
    _default_host_driver = "ec2"
    _default_host_driver_options = {
        'instanceType': None,
        'region': None, 
        'useSpotInstance': True,
        'maxSpotPrice': None
    }

    libraries = None  # type: Dict[str, str]
    """`Dict str=str`: A dictionary of `package=local_path` of local
            packages used inside the project from outside the 
            project's directory tree. These libraries will be 
            uploaded as part of your application and put on the
            `PYTHONPATH` during its execution."""

    memory = None     # type: int
    """`int`: The amount of memory in MB required by the job. Used
            to determine how many jobs will be run per host. Specifying
            an amount greater than the instance type specified in 
            `JobConfig.host_driver_options` (or the default instance type), will not
            change the instance type to accomodate the greater amount
            of memory, but will force only one job to run per host."""

    host_driver = None  # type: str
    """`str`: The backing driver to use in launching the hosts
            for the job. ("ec2" only currently)."""

    host_driver_options = None  # type: Dict[str, Any]
    """`Dict str=Any`: A dict of options for the host driver such
            as "instanceType", "region", and "useSpotInstance"."""

    host_affinity = None  # type: str
    """`str`: A string label used to group jobs on a host(s). If 
            specified, the job will only run on hosts with the same host
            affinity label. For example, assigning "host1" to jobs 1-4 and
            "host2" to jobs 5-8 will force both sets of jobs to run on 
            different hosts regardless of other factors."""

    capabilities = None  # type: List[aegisblade.capabilities.Capability]
    """`List` `aegisblade.capabilities.Capability`: An array of capabilities required by this job. See
            `aegisblade.capabilities.Capability` for more details."""

    def __init__(self):
        # type: (int, string, string, dict, list) -> None
        """Constructor intializes a `JobConfig` with default options."""
        self.libraries = {}
        self.memory = self._default_memory
        self.host_driver = Environment.default_hostdriver or self._default_host_driver
        self.host_driver_options = self._default_host_driver_options
        self.host_affinity = None
        self.capabilities = []

        env_libraries = Environment.libraries or dict()
        self.with_libraries(env_libraries)
    
    def with_memory(self, memory):
        # type: (int) -> JobConfig
        """Sets the amount of memory in MB required by the job.

        See the `JobConfig.memory` docs for more details on how memory is used.

        Args:
            memory (int): Amount of memory in MB.
        """
        if type(memory) is not int:
            raise TypeError('memory must be an int')

        self.memory = memory

        return self

    def with_host_driver(self, host_driver, host_driver_options=None):
        # type: (str, Optional[dict]) -> JobConfig
        """Sets the host driver and options for that driver to be used
        when starting the job.

        See the `JobConfig.host_driver` and `JobConfig.host_driver_options` docs for more detail.

        Args:
            host_driver (str): See `JobConfig.host_driver`

            host_driver_options (Dict str=Any): See `JobConfig.host_driver_options`

        """
        if type(host_driver) is not str:
            raise TypeError('host_driver must be a str')

        if host_driver_options is not None and type(host_driver_options) is not dict:
            raise TypeError('host_driver_options must be a dict')

        self.host_driver = host_driver

        if host_driver_options:
            self.host_driver_options = self._merge_default_dict(self._default_host_driver_options, host_driver_options)
        elif not self.host_driver_options:
            self.host_driver_options = self._default_host_driver_options

        return self

    def with_host_affinity(self, host_affinity):
        # type: (str) -> JobConfig
        """Sets the host affinity for the job.

        See the `JobConfig.host_affinity` doc for more details.

        Args:
            host_affinity (str): See `JobConfig.host_affinity`
        """
        if host_affinity is not None and type(host_affinity) is not str:
            raise TypeError('host_affinity must be a str')

        self.host_affinity = host_affinity

        return self

    def with_libraries(self, libraries):
        # type: (dict) -> JobConfig
        """Sets multiple libraries required by the job.

        See the `JobConfig.libraries` doc for more detail.

        Args:
            libraries (Dict str=str): See `JobConfig.libraries`
        """
        if not isinstance(libraries, dict):
            raise TypeError('libraries must be a dict with format {package_name=package_path}')

        for library_package, library_path in libraries.items():
            self.add_library(library_package, library_path)
        
        return self
    
    def with_capability(self, capability):
        # type: (str) -> JobConfig
        """Specifies a capability required by the job.

        See the `JobConfig.capabilities` doc for more details on capabilities.
        """
        if not isinstance(capability, basestring):
            raise TypeError('capability must a str-like type')

        self.capabilities.append(capability)

        return self

    def add_library(self, library_package, library_path):
        # type: (str) -> JobConfig
        """Specifies a library required by the job.

        See the `JobConfig.libraries` doc for more detail.

        Args:
            library_package (str): The name of the package defined by the library. 
            library_path (str): The local path to the library.
        """
        if not isinstance(library_package, basestring):
            raise TypeError('library_package must a str-like type')

        if not isinstance(library_path, basestring):
            raise TypeError('library_path must a str-like type')

        library_path = os.path.normpath(
            os.path.normcase(
                os.path.abspath(
                    os.path.expanduser(
                        library_path
                    ))))

        if not (os.path.exists(library_path)):
            raise ValueError(library_path + ' not found')

        if not os.path.isdir(library_path):
            raise ValueError(library_path + ' is not a directory')

        self.libraries[library_package] = library_path

        return self

    def _merge_default_dict(self, *source_dicts):
        merged = {}

        for source_dict in source_dicts:
            for key, value in source_dict.items():
                merged[key] = value
        
        return merged





# A part of the AegisBlade Python Client Library
# Copyright (C) 2019 Thornbury Organization, Bryan Thornbury
# This file may be used under the terms of the GNU Lesser General Public License, version 2.1.
# For more details see: https://www.gnu.org/licenses/lgpl-2.1.html

class InvalidApiKeyError(Exception):
    """An error raised when an invalid api key is provided."""

    def __init__(self, server_response_message):
        self.server_response_message = server_response_message

        self.message = "The server reported your api key is invalid. Check to ensure it's been entered correctly. " \
                       "\nServer Response: {0}".format(self.server_response_message)

    def __str__(self):
        return self.message


class ApplicationBuildError(Exception):
    """An error raised when an application fails to build."""

    def __init__(self, application_id, error_description):
        self.application_id = application_id
        self.error_description = error_description

        self.message = "An Application failed to build. File an issue at https://github.com/brthor/aegisblade if it persists. " \
                       "\n(Error Description: {0})" \
                       "\n(Application Id:{1})".format(self.error_description, self.application_id)

    def __str__(self):
        return self.message


class JobError(Exception):
    """An error raised when a job failed to begin running."""

    def __init__(self, job_id, error_description):
        self.job_id = job_id
        self.error_description = error_description

        self.message = "An Job failed to run. File an issue at https://github.com/brthor/aegisblade if it persists. " \
                       "\n(Error Description: {0})" \
                       "\n(Job Id:{1})".format(self.error_description, self.job_id)

    def __str__(self):
        return self.message


class JobRuntimeError(Exception):
    """An error raised when a job raises an Exception while running."""

    internal_error = None
    """The runtime error raised by the job while running."""

    def __init__(self, job_id, internal_error):
        # type: (str, Exception) -> None
        self.internal_error = internal_error
        self.job_id = job_id

        self.message = "A job raised an exception while running. See `internal_error` property for the exception." \
                       "\nInternal Error Type: {0}" \
                       "\nInternal Error: {1}" \
                       "\nJob Id: {2}".format(self.internal_error.__class__.__name__,
                                              str(self.internal_error),
                                              self.job_id)

    def __str__(self):
        return self.message


class UnableToDeserializeReturnValueError(Exception):
    """An error raised when a job returns a value which cannot be deserialized."""

    def __init__(self, job_id):
        self.job_id = job_id
        self.message = "A job returned a value which cannot be deserialized. " \
                       "\n(Job Id: {0})".format(self.job_id)

    def __str__(self):
        return self.message


class JobNotFoundError(Exception):
    """An error raised when a job with a specified id does not exist."""

    def __init__(self, job_id):
        self.job_id = job_id
        self.message = "A job with the specified id could not be located. " \
                       "\n(Job Id: {0})".format(self.job_id)

    def __str__(self):
        return self.message


class ApplicationNotFoundError(Exception):
    """An error raised when an application with a specified id does not exist."""

    def __init__(self, application_id):
        self.application_id = application_id
        self.message = "An application with the specified id could not be located. " \
                       "\n(Application Id: {0})".format(self.application_id)

    def __str__(self):
        return self.message



# A part of the AegisBlade Python Client Library
# Copyright (C) 2019 Thornbury Organization, Bryan Thornbury
# This file may be used under the terms of the GNU Lesser General Public License, version 2.1.
# For more details see: https://www.gnu.org/licenses/lgpl-2.1.html

__content__=["__future__",
"__main__",
"_dummy_thread",
"_thread",
"abc",
"aifc",
"argparse",
"array",
"ast",
"asynchat",
"asyncio",
"asyncore",
"atexit",
"audioop",
"base64",
"bdb",
"binascii",
"binhex",
"bisect",
"builtins",
"bz2",
"cProfile",
"calendar",
"cgi",
"cgitb",
"chunk",
"cmath",
"cmd",
"code",
"codecs",
"codeop",
"collections",
"collections.abc",
"colorsys",
"compileall",
"concurrent.futures",
"configparser",
"contextlib",
"copy",
"copyreg",
"crypt",
"csv",
"ctypes",
"curses",
"curses.ascii",
"curses.panel",
"curses.textpad",
"datetime",
"dbm",
"dbm.dumb",
"dbm.gnu",
"dbm.ndbm",
"decimal",
"difflib",
"dis",
"distutils",
"distutils.archive_util",
"distutils.bcppcompiler",
"distutils.ccompiler",
"distutils.cmd",
"distutils.command",
"distutils.command.bdist",
"distutils.command.bdist_dumb",
"distutils.command.bdist_msi",
"distutils.command.bdist_packager",
"distutils.command.bdist_rpm",
"distutils.command.bdist_wininst",
"distutils.command.build",
"distutils.command.build_clib",
"distutils.command.build_ext",
"distutils.command.build_py",
"distutils.command.build_scripts",
"distutils.command.check",
"distutils.command.clean",
"distutils.command.config",
"distutils.command.install",
"distutils.command.install_data",
"distutils.command.install_headers",
"distutils.command.install_lib",
"distutils.command.install_scripts",
"distutils.command.register",
"distutils.command.sdist",
"distutils.core",
"distutils.cygwinccompiler",
"distutils.debug",
"distutils.dep_util",
"distutils.dir_util",
"distutils.dist",
"distutils.errors",
"distutils.extension",
"distutils.fancy_getopt",
"distutils.file_util",
"distutils.filelist",
"distutils.log",
"distutils.msvccompiler",
"distutils.spawn",
"distutils.sysconfig",
"distutils.text_file",
"distutils.unixccompiler",
"distutils.util",
"distutils.version",
"doctest",
"dummy_threading",
"email",
"email.charset",
"email.contentmanager",
"email.encoders",
"email.errors",
"email.generator",
"email.header",
"email.headerregistry",
"email.iterators",
"email.message",
"email.mime",
"email.parser",
"email.policy",
"email.utils",
"encodings.idna",
"encodings.mbcs",
"encodings.utf_8_sig",
"ensurepip",
"enum",
"errno",
"faulthandler",
"fcntl",
"filecmp",
"fileinput",
"fnmatch",
"formatter",
"fpectl",
"fractions",
"ftplib",
"functools",
"gc",
"getopt",
"getpass",
"gettext",
"glob",
"grp",
"gzip",
"hashlib",
"heapq",
"hmac",
"html",
"html.entities",
"html.parser",
"http",
"http.client",
"http.cookiejar",
"http.cookies",
"http.server",
"imaplib",
"imghdr",
"imp",
"importlib",
"importlib.abc",
"importlib.machinery",
"importlib.util",
"inspect",
"io",
"ipaddress",
"itertools",
"json",
"json.tool",
"keyword",
"lib2to3",
"linecache",
"locale",
"logging",
"logging.config",
"logging.handlers",
"lzma",
"macpath",
"mailbox",
"mailcap",
"marshal",
"math",
"mimetypes",
"mmap",
"modulefinder",
"msilib",
"msvcrt",
"multiprocessing",
"multiprocessing.connection",
"multiprocessing.dummy",
"multiprocessing.managers",
"multiprocessing.pool",
"multiprocessing.sharedctypes",
"netrc",
"nis",
"nntplib",
"numbers",
"operator",
"optparse",
"os",
"os.path",
"ossaudiodev",
"parser",
"pathlib",
"pdb",
"pickle",
"pickletools",
"pipes",
"pkgutil",
"platform",
"plistlib",
"poplib",
"posix",
"pprint",
"profile",
"pstats",
"pty",
"pwd",
"py_compile",
"pyclbr",
"pydoc",
"queue",
"quopri",
"random",
"re",
"readline",
"reprlib",
"resource",
"rlcompleter",
"runpy",
"sched",
"secrets",
"select",
"selectors",
"shelve",
"shlex",
"shutil",
"signal",
"site",
"smtpd",
"smtplib",
"sndhdr",
"socket",
"socketserver",
"spwd",
"sqlite3",
"ssl",
"stat",
"statistics",
"string",
"stringprep",
"struct",
"subprocess",
"sunau",
"symbol",
"symtable",
"sys",
"sysconfig",
"syslog",
"tabnanny",
"tarfile",
"telnetlib",
"tempfile",
"termios",
"test",
"test.support",
"textwrap",
"threading",
"time",
"timeit",
"tkinter",
"tkinter.scrolledtext",
"tkinter.tix",
"tkinter.ttk",
"token",
"tokenize",
"trace",
"traceback",
"tracemalloc",
"tty",
"turtle",
"turtledemo",
"types",
"typing",
"unicodedata",
"unittest",
"unittest.mock",
"urllib",
"urllib.error",
"urllib.parse",
"urllib.request",
"urllib.response",
"urllib.robotparser",
"uu",
"uuid",
"venv",
"warnings",
"wave",
"weakref",
"webbrowser",
"winreg",
"winsound",
"wsgiref",
"wsgiref.handlers",
"wsgiref.headers",
"wsgiref.simple_server",
"wsgiref.util",
"wsgiref.validate",
"xdrlib",
"xml",
"xml.dom",
"xml.dom.minidom",
"xml.dom.pulldom",
"xml.etree.ElementTree",
"xml.parsers.expat",
"xml.parsers.expat.errors",
"xml.parsers.expat.model",
"xml.sax",
"xml.sax.handler",
"xml.sax.saxutils",
"xml.sax.xmlreader",
"xmlrpc.client",
"xmlrpc.server",
"zipapp",
"zipfile",
"zipimport",
"zlib"]



# A part of the AegisBlade Python Client Library
# Copyright (C) 2019 Thornbury Organization, Bryan Thornbury
# This file may be used under the terms of the GNU Lesser General Public License, version 2.1.
# For more details see: https://www.gnu.org/licenses/lgpl-2.1.html

from typing import List

from aegisblade._internal.application_package_info import ApplicationPackageInfo
from aegisblade._internal.application_context_file import ApplicationContextFile


class ApplicationExecutionContext(object):
    """Encapsulates the data required to build an executable 
    version of the application on AegisBlade.

    Attributes:
        language (str): The programming language used in the application. "python" in this case.

        sourceRuntimeIdentifier (str): OS Identier of the source machine. Used for diagnostic purposes.

        packages (List aegisblade.application_package_info.ApplicationPackageInfo): List of Application dependency packages, Name, Version pairs.

        files (List aegisblade.application_context_file.ApplicationContextFile): List of Application files which are required for the program to function as expected.

        targetRuntimeName (str): Not used in python.

        targetRuntimeVersion (str): Python version.

        packageManagerSourceFeeds (List str): Source feeds for packages used by your application. Not currently used in python.

        clientVersion (List): Version of this client.
    """

    packageManagerSourceFeeds = None    # type: List[str]
    targetRuntimeVersion = None         # type: str
    targetRuntimeName = None            # type: str
    files = None                        # type: List[ApplicationContextFile]
    packages = None                     # type: List[ApplicationPackageInfo]
    language = None                     # type: str
    sourceRuntimeIdentifier = None      # type: str
    clientVersion = None                # type: str

    def __init__(self,
                 language,
                 source_runtime_identifier,
                 packages,
                 files,
                 target_runtime_name,
                 target_runtime_version,
                 package_manager_source_feeds,
                 client_version):
        # type: (str, str, List[ApplicationPackageInfo], List[ApplicationContextFile], str, str, List[str], str) -> None
        """Initializes the ApplicationExecutionContext.

        Args:
            language (str): The programming language used in the application. "python" in this case.

            sourceRuntimeIdentifier (str): OS Identier of the source machine. Used for diagnostic purposes.

            packages (List aegisblade.application_package_info.ApplicationPackageInfo): List of Application dependency packages, Name, Version pairs.

            files (List aegisblade.application_context_file.ApplicationContextFile): List of Application files which are required for the program to function as expected.

            targetRuntimeName (str): Not used in python.

            targetRuntimeVersion (str): Python version.

            packageManagerSourceFeeds (List str): Source feeds for packages used by your application. Not currently used in python.

            clientVersion (List): Version of this client.
        """
        self.packageManagerSourceFeeds = package_manager_source_feeds
        self.targetRuntimeVersion = target_runtime_version
        self.targetRuntimeName = target_runtime_name
        self.files = files
        self.packages = packages
        self.language = language
        self.sourceRuntimeIdentifier = source_runtime_identifier
        self.clientVersion = client_version




# A part of the AegisBlade Python Client Library
# Copyright (C) 2019 Thornbury Organization, Bryan Thornbury
# This file may be used under the terms of the GNU Lesser General Public License, version 2.1.
# For more details see: https://www.gnu.org/licenses/lgpl-2.1.html

from typing import List


class CreateApplicationResponse(object):
    """Parsed API response when creating an application."""

    error = None  # type: bool
    """`bool`: Set to True if an error occurred."""

    errorDescription = None  # type: str
    """`str`: Describes the error encountered when creating the application, if one occurred."""

    applicationId = None  # type: str
    """`str`: The GUID-like id of the application."""

    fileHashesRequiringUpload = None  # type: List[str]
    """`List[str]`: A list of hashes of files requiring upload before the application can be built."""

    def __str__(self):
        return self.__dict__.__str__()


class CreateJobResponse(object):
    """Parsed API response when creating an job."""

    error = None  # type: bool
    """`bool`: Set to True if an error occurred."""

    errorDescription = None  # type: str
    """`str`: Describes the error encountered when creating the job, if one occurred."""

    jobId = None  # type: str
    """`str`: The GUID-like id of the job."""

    jobType = None  # type: str
    """`str`: The type of job that was created."""

    def __str__(self):
        return self.__dict__.__str__()


class JobStatusResponse(object):
    """Parsed API response when fetching a job's status."""

    jobId = None  # type: str
    """`str`: The GUID-like id of the job."""

    jobStatus = None  # type: str
    """`str`: The status of the job."""

    jobType = None  # type: str
    """`str`: The type of the job."""

    errorDescription = None  # type: str
    """`str`: Describes the error encountered when running the job, if one occurred."""

    applicationId = None  # type: str
    """`str`: The GUID-like id of the application the job is a part of."""

    def __str__(self):
        return self.__dict__.__str__()


class ApplicationStatusResponse(object):
    """Parsed API response when fetching an application's status."""

    applicationId = None  # type: str
    """`str`: The GUID-like id of the application."""

    status = None  # type: str
    """`str`: The status of the job."""

    errorDescription = None  # type: str
    """`str`: Describes the error encountered when building the application, if one occurred."""

    def __str__(self):
        return self.__dict__.__str__()



# A part of the AegisBlade Python Client Library
# Copyright (C) 2019 Thornbury Organization, Bryan Thornbury
# This file may be used under the terms of the GNU Lesser General Public License, version 2.1.
# For more details see: https://www.gnu.org/licenses/lgpl-2.1.html

import base64
import sys
import cloudpickle
from typing import Callable


class TaskInfo(object):
    """Wraps and serializes the callable to be run as part of a job.

    Attributes:
        func: The callable to be run as part of a job.
    """

    func = None  # type: Callable

    def __init__(self, func):
        # type: (Callable) -> None
        self.func = func

    def serializeForInjector(self):
        # type: () -> str
        ret = cloudpickle.dumps(self.func)
        ret = base64.b64encode(ret)
        
        if sys.version_info[0] == 3:
            ret = ret.decode('ascii')
            
        return ret



# A part of the AegisBlade Python Client Library
# Copyright (C) 2019 Thornbury Organization, Bryan Thornbury
# This file may be used under the terms of the GNU Lesser General Public License, version 2.1.
# For more details see: https://www.gnu.org/licenses/lgpl-2.1.html

import shutil


class DataStore(object):
    """A utility for accessing data stores and their files.

    A Data Store is a logical "drive" of files that is backed
    by your choice of driver (AWS S3 only currently). Before
    uploading files or other operations the data store must be
    created with a call to `DataStore.create()`.

    Each data store is identified by a unique name, and can 
    be accessed from any job or machine with access to the proper
    API Key. 

    The data store is typically used to save results from jobs or 
    put up data that can be accessed by any of your jobs.

    Attributes:
        name (str): The name of the data store.

    """

    name = None
    """`str`: The name of the data store."""


    def __init__(self, name, api_consumer):
        # type: (str, AegisBladeApiConsumer) -> None
        """The constructor is not recommended for direct use. 
        Use `aegisblade.client.AegisBladeClient.data()` to get a `DataStore` instance.

        Call `DataStore.create()` to create the data store if it does
        not already exist.
        """

        self.name = name
        """`str`: The name of the data store."""

        self.api_consumer = api_consumer

    def create(self, driver, driver_options):
        # type: (string, dict) -> DataStore
        """Creates the data store using the specified backing driver
        and options for that driver.

        Args:
            driver (str): The driver to use for backing the data store ("s3" only currently.)

            driver_options (dict): A dictionary of options specific to the driver. 
                (e.g. "region" may be defined for s3)
        """
        self.api_consumer.data_store_create(self.name, driver, driver_options)

        return self

    def upload_file(self, local_file_path, data_store_path):
        # type: (str, str) -> None
        """Uploads a file from the local machine to the data store.
        
        Args:
            local_file_path (str): The path of the file to upload on the local machine.

            data_store_path (str): The storage path of the file inside the data store.
        """
        with open(local_file_path, 'rb') as f:
            self.api_consumer.data_store_upload(self.name, data_store_path, f)

    def upload_data(self, data, data_store_path):
        # type: (str, str) -> None
        """Uploads a string to the data store.
        
        Args:
            data (str): The data string to store in the data store.

            data_store_path (str): The path in which to store the data inside the data store.
        """
        self.api_consumer.data_store_upload(self.name, data_store_path, data)

    def download(self, data_store_path):
        # type: (str) -> str
        """Downloads data from the data store and returns it as a string.

        Args:
            data_store_path (str): The path of the data inside the data store to download.

        Returns:
            `str`: The downloaded data as a string.

            On Python3 this will return a `bytes` object.
        """
        tmp_file = self.api_consumer.data_store_download(self.name, data_store_path)
        
        with open(tmp_file, 'rb') as f:
            return f.read()

    def download_to_file(self, data_store_path, local_file):
        # type: (str, str) -> None
        """Downloads data from the data store and writes it to a local file.

        Args:
            data_store_path (str): The path of the data inside the data store to download.
            
            local_file (str): Path of a local file that the data will be written to.
        """
        tmp_file = self.api_consumer.data_store_download(self.name, data_store_path)

        shutil.copyfile(tmp_file, local_file)

    def delete(self, data_store_path):
        # type: (str) -> None
        """Deletes the specified data from the data store.

        Args:
            data_store_path (str): Path of the data inside the data store to delete.
        """
        self.api_consumer.data_store_delete_file(self.name, data_store_path)

    def delete_store(self):
        """Deletes this entire data store and all files in it.

        USE CAREFULLY!
        """
        self.api_consumer.data_store_delete(self.name)

    def list_files(self):
        """Lists all files in the data store.

        Returns:
            `List str`: A list of all files in the data store.
        """
        return self.api_consumer.data_store_list_files(self.name)



# A part of the AegisBlade Python Client Library
# Copyright (C) 2019 Thornbury Organization, Bryan Thornbury
# This file may be used under the terms of the GNU Lesser General Public License, version 2.1.
# For more details see: https://www.gnu.org/licenses/lgpl-2.1.html

class ApplicationPackageInfo(object):
    """Stores information related to a 'pip install'ed package
    in the current environment.

    Attributes:
        name (str): Name of the package

        version (str): Full version of the package.
    """

    name = None     # type: str
    version = None  # type: str

    def __init__(self, name, version):
        """Initializes the ApplicationPackageInfo.

        Args:
            name (str): Name of the package

            version (str): Full version of the package.
        """
        self.name = name
        self.version = version


# A part of the AegisBlade Python Client Library
# Copyright (C) 2019 Thornbury Organization, Bryan Thornbury
# This file may be used under the terms of the GNU Lesser General Public License, version 2.1.
# For more details see: https://www.gnu.org/licenses/lgpl-2.1.html

from aegisblade.ApiModels.responses import ApplicationStatusResponse
from aegisblade._internal.api_consumer import _AegisBladeApiConsumer


class AegisBladeApplication(object):
    """An object for fetching and storing application-related information 
    from the API.

    This is typically returned by a call to `aegisblade.client.AegisBladeClient.run()` but may
    be constructed from a saved `AegisBladeApplication.id` by calling the 
    `AegisBladeApplication.create()` method.

    In AegisBlade, an application is the collection of files, packages, 
    libraries, and capabilities that constitute an executable codebase.
    
    These constituents are collected and constructed into a deployable image
    during the application build. Then any number of jobs can be defined to
    run on top of this application image.

    Attributes:
        id (str): The GUID-like id of the application.
    """

    id = None               # type: str
    """`str`: The GUID-like id of the application."""

    _api_consumer = None     # type: AegisBladeApiConsumer

    @classmethod
    def create(cls, client, application_id):
        # type: (AegisBladeClient, str)
        """Creates an AegisBladeApplication instance from an application id.

        Args:
            client (aegisblade.client.AegisBladeClient): The AegisBlade client. The default client is imported 
                using 'from aegisblade import aegisblade'.

            application_id (str): The id of the application corresponding to the 
                `AegisBladeApplication.id` instance property.
        """
        return cls(client._get_api_consumer(), application_id)

    def __init__(self, api_consumer, application_id):
        # type: (AegisBladeApiConsumer, str) -> None
        self._api_consumer = api_consumer
        self.id = application_id

    def get_status(self):
        # type: () -> ApplicationStatusResponse
        """Gets the status information of the application from the API.

        Returns:
            aegisblade.ApiModels.responses.ApplicationStatusResponse
        """
        return self._api_consumer.application_status(self.id)



# A part of the AegisBlade Python Client Library
# Copyright (C) 2019 Thornbury Organization, Bryan Thornbury
# This file may be used under the terms of the GNU Lesser General Public License, version 2.1.
# For more details see: https://www.gnu.org/licenses/lgpl-2.1.html

"""Package `aegisblade` provides the Python client library for [AegisBlade](/).

This is the reference documentation for the Python client library.

[Read the docs](/docs) for more detail.

[See more examples](https://www.github.com/aegisblade/examples) on Github.

Installation
=============

```pip install aegisblade```


Examples
==========

#### Hello World

```
from aegisblade import aegisblade

aegisblade.set_api_key("MY-API-KEY")

def hello_world():
    print("Hello World")
    return "Hello World"

job = aegisblade.run(lambda: do_work())

assert "Hello World" in job.get_return_value()
```

"""
from aegisblade.client import AegisBladeClient
from aegisblade.job_config import JobConfig
from aegisblade.capabilities import Capability

from aegisblade.__version__ import __version__

aegisblade = AegisBladeClient()
__version__ = __version__



# A part of the AegisBlade Python Client Library
# Copyright (C) 2019 Thornbury Organization, Bryan Thornbury
# This file may be used under the terms of the GNU Lesser General Public License, version 2.1.
# For more details see: https://www.gnu.org/licenses/lgpl-2.1.html

import requests, shutil, functools, os, tempfile
from requests import HTTPError
from requests.utils import super_len

from aegisblade.ApiModels.payloads import CreateJobPayload, UploadFilePayload, CreateApplicationPayload, CreateDataStorePayload
from aegisblade.ApiModels.responses import JobStatusResponse, CreateJobResponse, CreateApplicationResponse, \
    ApplicationStatusResponse
from aegisblade._internal.application_context_file import ApplicationContextFile
from aegisblade._internal.application_execution_context import ApplicationExecutionContext
from aegisblade.errors import InvalidApiKeyError
from aegisblade.job_config import JobConfig
from aegisblade._internal.task_info import TaskInfo
from aegisblade._internal.json_wrapper import json_encode, json_decode
from aegisblade.__version__ import __version__


class _AegisBladeApiConsumer(object):
    """Internal class that abstracts the HTTP interface of the AegisBlade API.
    """
    def __init__(self, api_key, service_url="https://localhost:5000/api/v1", verify_ssl=True):
        self.service_url = service_url
        self.api_key = api_key
        self.verify_ssl = verify_ssl

    def create_application(self, application_execution_context, job_config):
        # type: (ApplicationExecutionContext, JobConfig) -> CreateApplicationResponse
        url = self.service_url + "/application/create"

        post_data = CreateApplicationPayload(application_execution_context, job_config)
        post_data = json_encode(post_data)

        response = requests.post(url, data=post_data, headers=self._get_default_headers(), verify=self.verify_ssl)
        self._handle_potential_error(response)

        application_response = json_decode(response.text, CreateApplicationResponse)

        return application_response

    def application_status(self, application_guid):
        # type: (str) -> ApplicationStatusResponse
        '''
        :type application_guid: str
        :rtype: ApplicationStatusResponse
        '''
        url = self.service_url + "/application/status/" + application_guid

        response = requests.get(url, headers=self._get_default_headers(), verify=self.verify_ssl)
        self._handle_potential_error(response)

        job_response = json_decode(response.text, ApplicationStatusResponse)
        return job_response

    def upload_file(self, application_id, application_context_file):
        # type: (str, ApplicationContextFile) -> str
        url = self.service_url + "/application/upload"

        post_data = UploadFilePayload(application_context_file, application_context_file._x_fileContents, application_id)
        post_data = json_encode(post_data)

        response = requests.post(url, data=post_data, headers=self._get_default_headers(), verify=self.verify_ssl)
        self._handle_potential_error(response)

        return response.text

    def create_instant_job(self, task_info, application_id, job_config):
        # type: (TaskInfo, str, JobConfig) -> CreateJobResponse
        url = self.service_url + "/job/create"

        post_data = CreateJobPayload(task_info, application_id, "instant",
                                    __version__, job_config)
        post_data = json_encode(post_data)

        response = requests.post(url, data=post_data, headers=self._get_default_headers(), verify=self.verify_ssl)
        self._handle_potential_error(response)

        job_response = json_decode(response.text, CreateJobResponse)

        return job_response

    def job_status(self, job_guid):
        # type: (str) -> JobStatusResponse
        url = self.service_url + "/job/status/" + job_guid

        response = requests.get(url, headers=self._get_default_headers(), verify=self.verify_ssl)
        self._handle_potential_error(response)

        job_response = json_decode(response.text, JobStatusResponse)
        return job_response

    def _get_default_headers(self, content_type="application/json"):
        return {
            "Content-Type": content_type,
            "Authorization": "bearer " + self.api_key
        }

    def job_logs(self, job_guid):
        # type: (str) -> str
        url = self.service_url + "/job/logs/" + job_guid

        response = requests.get(url, headers=self._get_default_headers(), verify=self.verify_ssl)
        self._handle_potential_error(response)

        return response.text

    def job_return_value(self, job_guid):
        # type: (str) -> str
        url = self.service_url + "/job/returnvalue/" + job_guid

        response = requests.get(url, headers=self._get_default_headers(), verify=self.verify_ssl)
        self._handle_potential_error(response)

        return response.text

    def data_store_create(self, data_store_name, data_storage_driver, driver_options):
        url = self.service_url + "/data/store/create"

        post_data = CreateDataStorePayload(data_store_name, data_storage_driver, driver_options)
        post_data = json_encode(post_data)

        response = requests.post(url, data=post_data, headers=self._get_default_headers(), verify=self.verify_ssl)
        self._handle_potential_error(response)

        return response.text

    def data_store_delete(self, data_store_name):
        url = self.service_url + "/data/store/{0}/delete".format(data_store_name)

        response = requests.delete(url, headers=self._get_default_headers(), verify=self.verify_ssl)
        self._handle_potential_error(response)

        return response.text

    def data_store_upload(self, data_store_name, data_store_path, data):
        url = self.service_url + "/data/store/{0}/file/{1}".format(data_store_name, data_store_path)

        response = requests.put(url, data=data, 
            headers=self._get_default_headers(content_type="application/octet-stream"), 
            verify=self.verify_ssl,
            allow_redirects=False)
        self._handle_potential_error(response)
        
        # Redirects are expected
        if response.status_code != 302:
            return
        
        redirect_url = response.headers.get('location')

        if hasattr(data, 'seek'):
            data.seek(0)

        response2 = requests.put(redirect_url, data=data, 
            headers={'Content-Type': "application/octet-stream"}, 
            verify=True,
            allow_redirects=True,
            stream=True)
        
        self._handle_potential_error(response2)


    def data_store_delete_file(self, data_store_name, data_store_path):
        url = self.service_url + "/data/store/{0}/file/{1}".format(data_store_name, data_store_path)

        response = requests.delete(url, 
            headers=self._get_default_headers(), 
            verify=self.verify_ssl)

        self._handle_potential_error(response)

        return response.text

    def data_store_download(self, data_store_name, data_store_path):
        url = self.service_url + "/data/store/{0}/file/{1}".format(data_store_name, data_store_path)

        temp_file = os.path.join(tempfile.mkdtemp(), os.path.basename(data_store_path))

        redirect_url = self._stream_download_file(url, 
            temp_file, self._get_default_headers(), self.verify_ssl, False)

        # Storage driver specific url redirections are expected
        if redirect_url:
            self._stream_download_file(redirect_url, temp_file, None, True, True)

        return temp_file

    def data_store_list_files(self, data_store_name):
        url = self.service_url + "/data/store/{0}/list".format(data_store_name)

        response = requests.get(url, headers = self._get_default_headers(), verify=self.verify_ssl)
        self._handle_potential_error(response)

        json_response = response.json()

        return json_response

    def _stream_download_file(self, url, local_filename, headers, verify_ssl, allow_redirects):
        with requests.get(url, stream=True, headers=headers, verify=verify_ssl, allow_redirects=allow_redirects) as r:
            if (r.status_code == 302):
                return r.headers.get('location')

            r.raise_for_status()

            r.raw.read = functools.partial(r.raw.read, decode_content=True)

            with open(local_filename, 'wb') as f:
                shutil.copyfileobj(r.raw, f)

        return None

    def _handle_potential_error(self, response):
        # type: (requests.Response) -> None

        # Authentication Error
        if response.status_code == 401 and response.headers.get('www-authenticate', None) is not None:
            raise InvalidApiKeyError(
                "The server reported your api key is invalid. Check to ensure it's been entered correctly. "
                "Server Response: {0}".format(response.headers.get('www-authenticate')))

        # Any kind of error where the server returns a json message
        if response.status_code not in [200, 201] \
                and response.headers.get("content-type", "").startswith("application/json"):
            try:
                response_json = response.json()
                error_message = response_json.get("errorMessage", None)
            except ValueError as e:
                if 'json' not in str(e).lower():
                    raise

                error_message = None

            if error_message is not None:
                raise HTTPError(error_message, response=response)

        # Catch-All
        response.raise_for_status()




# A part of the AegisBlade Python Client Library
# Copyright (C) 2019 Thornbury Organization, Bryan Thornbury
# This file may be used under the terms of the GNU Lesser General Public License, version 2.1.
# For more details see: https://www.gnu.org/licenses/lgpl-2.1.html

import subprocess

from typing import List


class Command(object):
    """Internal class for executing Processes"""
    executable = None
    arguments = None

    def __init__(self, executable, arguments):
        # type: (str, List[str]) -> None
        self.executable = executable
        self.arguments = arguments

        self.environment_variables = dict()

    def environment(self, key, value):
        self.environment_variables[key] = value
        return self

    def environment_dict(self, env_dict):
        for key, value in env_dict.items():
            self.environment_variables[key] = value
        return self

    def execute(self):
        s_process_args = [self.executable]
        s_process_args.extend(self.arguments)

        s_process = subprocess.Popen(
            s_process_args,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=self.environment_variables,
            universal_newlines=True)

        out, err = s_process.communicate()
        exitcode = s_process.returncode

        # print out
        # print err

        return ProcessResult(self, out, err, exitcode)


class ProcessResult():
    """
    :type process: Command
    :type s_process: subprocess.Popen
    :type stdout: str
    :type std
    """

    def __init__(self, process, stdout, stderr, exitcode):
        self.process = process
        self.stdout = stdout
        self.stderr = stderr
        self.exitcode = exitcode

    def failed(self):
        return self.exitcode != 0



# A part of the AegisBlade Python Client Library
# Copyright (C) 2019 Thornbury Organization, Bryan Thornbury
# This file may be used under the terms of the GNU Lesser General Public License, version 2.1.
# For more details see: https://www.gnu.org/licenses/lgpl-2.1.html

import base64
import inspect

import time
from aegisblade._internal.atrace import Trace

import os
import sys

import pickle
from typing import Callable, List

from aegisblade._internal.api_consumer import _AegisBladeApiConsumer
from aegisblade._internal.application_execution_context_collector import _ApplicationExecutionContextCollector
from aegisblade._internal.task_info import TaskInfo
from aegisblade.job import AegisBladeJob
from aegisblade.job_config import JobConfig
from aegisblade.data_store import DataStore

from aegisblade._internal.environment import Environment

class AegisBladeClient(object):
    """AegisBlade client. Provides easy methods for starting jobs,
    and accessing data files.

    The client may be configured by using its methods such as `AegisBladeClient.set_api_key()`
    or by the use of the following environment variables:

    Environment:

        AEGISBLADE_API_KEY: Your api key.

        AEGISBLADE_DEFAULT_HOSTDRIVER: The default hostdriver to use such as 'ec2'.

        AEGISBLADE_VERIFY_SSL: Whether or not to verify the ssl certs of the endpoint. Testing purposes only.

        AEGISBLADE_API_ENDPOINT: The service endpoint. Testing purposes only.

        AEGISBLADE_DEBUG_OUTPUT: Enable debug console output.

    """
    _api_key = None
    _service_url = None

    _verify_ssl = True

    _default_endpoint = "https://www.aegisblade.com"
    _default_service_path = "/api/v1"

    def __init__(self, verify_ssl=True):
        """Initializes the AegisBlade client and uses the environment
        variables described in the class documentation to set values.
        """
        self.set_api_key(Environment.api_key)
        self.set_api_endpoint(Environment.api_endpoint or self._default_endpoint)
        
        env_verify_ssl = Environment.verify_ssl
        if env_verify_ssl is not None:
            self.set_verify_ssl(env_verify_ssl)

    def set_verify_ssl(self, verify_ssl):
        # type: (bool) -> None
        """Sets whether or not to verify the ssl certificates of the API endpoint.
        Not recommended outside of internal use.
        """
        if not isinstance(verify_ssl, bool):
            raise TypeError("Invalid verify_ssl type, expected bool.")

        self._verify_ssl = verify_ssl

    def set_api_key(self, api_key):
        """Sets the api key used by this client to access the AegisBlade API."""
        self._api_key = api_key

    def set_api_endpoint(self, endpoint):
        """Sets the api endpoint used by this client. 
        Not recommended outside of internal use.
        """
        if not endpoint:
            raise ValueError("Invalid endpoint value.")

        self.api_endpoint = endpoint

        self.set_service_url(self.api_endpoint + self._default_service_path)

    def set_service_url(self, service_url):
        """Sets the full service url (e.g. endpoint + path) used by this client. 
        Not recommended outside of internal use.
        """
        self._service_url = service_url

    def data(self, data_store_name):
        # type: (string) -> DataStore
        """Used to access the data store api. The data store can be used to 
        store arbitrary data accessible by any job or machine with internet 
        access (and the proper api key).

        Args:
            data_store_name (str): The name of the data store you want to access. 
                If it is not already created, you will need to call `aegisblade.data_store.DataStore.create()` 
                on the return value of this function.

        Returns:
            `aegisblade.data_store.DataStore`: A `aegisblade.data_store.DataStore` object which acts as a container of data files that you 
            may upload, download, or delete using the object. See `aegisblade.data_store.DataStore` class
            docs for more detail. 

        """
        return DataStore(data_store_name, self._get_api_consumer())

    def run(self, remotely_executed_callable, job_config=None):
        # type: (Callable, JobConfig) -> AegisBladeJob
        """Creates and runs a job on AegisBlade.
        
        When called this function will first create an application by uploading
        .py files found inside the current working directory tree, determine 
        libraries by calling 'pip freeze', and upload local libraries explicitly
        defined by the job_config argument.

        Refer to the `AegisBladeClient.get_upload_files()` method if you wish to 
        preview the files that will be uploaded, prior to running a job.

        After the files are uploaded successfully, AegisBlade will build an 
        application image that can be deployed and run on any number of machines.
        Jobs created as a part of this application will need to wait for it to finish
        building before running. The length of the build depends primarily upon the
        time required to install the various dependencies of the application.

        Applications are cached by AegisBlade, and subsequent calls to this function
        will only upload updated files. If no files have been updated, the same 
        application will be used, and created jobs will not need to wait for it to build.

        This function will queue a job to be run as a part of the application immediately
        after creating it and uploading the files. 
        
        Any number of jobs may be queued for an application, and there is no need to wait 
        for an application to be finished building or for the first job to run before 
        queuing more jobs.

        The returned object of this function may be used to check on the status of the job
        or application, or you can use the web ui available at https://www.aegisblade.com/app.

        Args:
            remotely_executed_callable (Callable): The target function you wish to execute.

            job_config (aegisblade.job_config.JobConfig): An optional JobConfig object used to configure
                various properties of the job such as required memory, and local libraries. 
                See JobConfig class documentation for more details.

        Returns:
            An `aegisblade.job.AegisBladeJob` class instance. This class may be used to fetch the job status,
            synchronously wait for it to finish, fetch logs, and fetch the return value.

            The `aegisblade.job.AegisBladeJob.id` property of the AegisBladeJob instance may be saved and used to get 
            information on the job at a later time.
        """
        if not callable(remotely_executed_callable):
            raise Exception("Must pass callable as first argument to AegisBlade.")

        if not self._api_key:
            raise Exception("The api key was not specified. Call `.set_api_key()` or set the AEGISBLADE_API_KEY environment variable.")


        job_config = job_config or JobConfig()

        task_info = TaskInfo(remotely_executed_callable)
        application_execution_context = _ApplicationExecutionContextCollector().collect(job_config.libraries)

        api_consumer = self._get_api_consumer()

        create_application_response = self._create_application(api_consumer, application_execution_context, job_config)

        self._upload_application_files(api_consumer, application_execution_context, create_application_response)
        create_job_response = self._create_job(api_consumer, task_info, create_application_response, job_config)

        return AegisBladeJob(api_consumer, create_job_response.jobId, create_application_response.applicationId, create_job_response.jobType)

    def get_upload_files(self, job_config):
        # type: (JobConfig) -> List[str]
        """Returns a list of files that will be uploaded as part of the
        application build. It DOES NOT upload any files or contact the API.

        This method exists so developers can check what files are going
        to be uploaded when they call `AegisBladeClient.run()`. 
        """
        job_config = job_config or JobConfig()

        application_execution_context = _ApplicationExecutionContextCollector().collect(job_config.libraries)

        local_upload_paths = []

        for application_file in application_execution_context.files:
            local_upload_paths.append(application_file.filePath)

        return local_upload_paths

    def _get_api_consumer(self):
        # type: () -> AegisBladeApiConsumer
        return _AegisBladeApiConsumer(self._api_key, service_url=self._service_url, verify_ssl=self._verify_ssl)

    def _create_application(self, api_consumer, application_execution_context, job_config):
        # type: (AegisBladeApiConsumer, ApplicationExecutionContext, JobConfig) -> CreateApplicationResponse
        create_application_response = api_consumer.create_application(application_execution_context, job_config)

        if create_application_response.error:
            raise Exception("Failed to create application. Error: {0}".format(create_application_response.errorDescription))

        Trace.verbose("[INFO] Application Created ({0}).".format(create_application_response.applicationId))

        return create_application_response

    def _upload_application_files(self, api_consumer, application_execution_context, create_application_response):
        # type: (AegisBladeApiConsumer, ApplicationExecutionContext, CreateApplicationResponse) -> None
        for count, upload_file_hash in enumerate(create_application_response.fileHashesRequiringUpload):
            app_file = [x for x in application_execution_context.files if x.fileHash == upload_file_hash][0]

            Trace.verbose("[INFO] Uploading File {0}/{1} ({2})"
                          .format(count+1,
                                  len(create_application_response.fileHashesRequiringUpload),
                                  app_file.filePathRelativeToAppContext))

            api_consumer.upload_file(create_application_response.applicationId, app_file)

    def _create_job(self, api_consumer, task_info, create_application_response, job_config):
        # type: (AegisBladeApiConsumer, TaskInfo, CreateApplicationResponse, JobConfig) -> CreateJobResponse
        Trace.verbose("[INFO] Creating Job.")
        create_job_response = api_consumer.create_instant_job(task_info, create_application_response.applicationId, job_config)

        if create_job_response.error:
            raise Exception("Failed to create job. Error: {0}".format(create_job_response.errorDescription))

        Trace.verbose("[INFO] Job Created ({0})".format(create_job_response.jobId))
        return create_job_response



# A part of the AegisBlade Python Client Library
# Copyright (C) 2019 Thornbury Organization, Bryan Thornbury
# This file may be used under the terms of the GNU Lesser General Public License, version 2.1.
# For more details see: https://www.gnu.org/licenses/lgpl-2.1.html




# A part of the AegisBlade Python Client Library
# Copyright (C) 2019 Thornbury Organization, Bryan Thornbury
# This file may be used under the terms of the GNU Lesser General Public License, version 2.1.
# For more details see: https://www.gnu.org/licenses/lgpl-2.1.html

__content__=["AL",
"BaseHTTPServer",
"Bastion",
"CGIHTTPServer",
"Carbon.AE",
"Carbon.AH",
"Carbon.App",
"Carbon.Appearance",
"Carbon.CF",
"Carbon.CG",
"Carbon.CarbonEvents",
"Carbon.CarbonEvt",
"Carbon.Cm",
"Carbon.Components",
"Carbon.ControlAccessor",
"Carbon.Controls",
"Carbon.CoreFounation",
"Carbon.CoreGraphics",
"Carbon.Ctl",
"Carbon.Dialogs",
"Carbon.Dlg",
"Carbon.Drag",
"Carbon.Dragconst",
"Carbon.Events",
"Carbon.Evt",
"Carbon.File",
"Carbon.Files",
"Carbon.Fm",
"Carbon.Folder",
"Carbon.Folders",
"Carbon.Fonts",
"Carbon.Help",
"Carbon.IBCarbon",
"Carbon.IBCarbonRuntime",
"Carbon.Icns",
"Carbon.Icons",
"Carbon.Launch",
"Carbon.LaunchServices",
"Carbon.List",
"Carbon.Lists",
"Carbon.MacHelp",
"Carbon.MediaDescr",
"Carbon.Menu",
"Carbon.Menus",
"Carbon.Mlte",
"Carbon.OSA",
"Carbon.OSAconst",
"Carbon.QDOffscreen",
"Carbon.Qd",
"Carbon.Qdoffs",
"Carbon.Qt",
"Carbon.QuickDraw",
"Carbon.QuickTime",
"Carbon.Res",
"Carbon.Resources",
"Carbon.Scrap",
"Carbon.Snd",
"Carbon.Sound",
"Carbon.TE",
"Carbon.TextEdit",
"Carbon.Win",
"Carbon.Windows",
"ColorPicker",
"ConfigParser",
"Cookie",
"DEVICE",
"DocXMLRPCServer",
"EasyDialogs",
"FL",
"FrameWork",
"GL",
"HTMLParser",
"MacOS",
"MimeWriter",
"MiniAEFrame",
"Nav",
"PixMapWrapper",
"Queue",
"SUNAUDIODEV",
"ScrolledText",
"SimpleHTTPServer",
"SimpleXMLRPCServer",
"SocketServer",
"StringIO",
"Tix",
"Tkinter",
"UserDict",
"UserList",
"UserString",
"W",
"__builtin__",
"__future__",
"__main__",
"_winreg",
"abc",
"aepack",
"aetools",
"aetypes",
"aifc",
"al",
"anydbm",
"applesingle",
"argparse",
"array",
"ast",
"asynchat",
"asyncore",
"atexit",
"audioop",
"autoGIL",
"base64",
"bdb",
"binascii",
"binhex",
"bisect",
"bsddb",
"buildtools",
"bz2",
"cPickle",
"cProfile",
"cStringIO",
"calendar",
"cd",
"cfmfile",
"cgi",
"cgitb",
"chunk",
"cmath",
"cmd",
"code",
"codecs",
"codeop",
"collections",
"colorsys",
"commands",
"compileall",
"compiler",
"compiler.ast",
"compiler.visitor",
"contextlib",
"cookielib",
"copy",
"copy_reg",
"crypt",
"csv",
"ctypes",
"curses",
"curses.ascii",
"curses.panel",
"curses.textpad",
"datetime",
"dbhash",
"dbm",
"decimal",
"difflib",
"dircache",
"dis",
"distutils",
"distutils.archive_util",
"distutils.bcppcompiler",
"distutils.ccompiler",
"distutils.cmd",
"distutils.command",
"distutils.command.bdist",
"distutils.command.bdist_dumb",
"distutils.command.bdist_msi",
"distutils.command.bdist_packager",
"distutils.command.bdist_rpm",
"distutils.command.bdist_wininst",
"distutils.command.build",
"distutils.command.build_clib",
"distutils.command.build_ext",
"distutils.command.build_py",
"distutils.command.build_scripts",
"distutils.command.check",
"distutils.command.clean",
"distutils.command.config",
"distutils.command.install",
"distutils.command.install_data",
"distutils.command.install_headers",
"distutils.command.install_lib",
"distutils.command.install_scripts",
"distutils.command.register",
"distutils.command.sdist",
"distutils.core",
"distutils.cygwinccompiler",
"distutils.debug",
"distutils.dep_util",
"distutils.dir_util",
"distutils.dist",
"distutils.emxccompiler",
"distutils.errors",
"distutils.extension",
"distutils.fancy_getopt",
"distutils.file_util",
"distutils.filelist",
"distutils.log",
"distutils.msvccompiler",
"distutils.spawn",
"distutils.sysconfig",
"distutils.text_file",
"distutils.unixccompiler",
"distutils.util",
"distutils.version",
"dl",
"doctest",
"dumbdbm",
"dummy_thread",
"dummy_threading",
"email",
"email.charset",
"email.encoders",
"email.errors",
"email.generator",
"email.header",
"email.iterators",
"email.message",
"email.mime",
"email.parser",
"email.utils",
"encodings.idna",
"encodings.utf_8_sig",
"ensurepip",
"errno",
"exceptions",
"fcntl",
"filecmp",
"fileinput",
"findertools",
"fl",
"flp",
"fm",
"fnmatch",
"formatter",
"fpectl",
"fpformat",
"fractions",
"ftplib",
"functools",
"future_builtins",
"gc",
"gdbm",
"gensuitemodule",
"getopt",
"getpass",
"gettext",
"gl",
"glob",
"grp",
"gzip",
"hashlib",
"heapq",
"hmac",
"hotshot",
"hotshot.stats",
"htmlentitydefs",
"htmllib",
"httplib",
"ic",
"icopen",
"imageop",
"imaplib",
"imgfile",
"imghdr",
"imp",
"importlib",
"imputil",
"inspect",
"io",
"itertools",
"jpeg",
"json",
"keyword",
"lib2to3",
"linecache",
"locale",
"logging",
"logging.config",
"logging.handlers",
"macerrors",
"macostools",
"macpath",
"macresource",
"mailbox",
"mailcap",
"marshal",
"math",
"md5",
"mhlib",
"mimetools",
"mimetypes",
"mimify",
"mmap",
"modulefinder",
"msilib",
"msvcrt",
"multifile",
"multiprocessing",
"multiprocessing.connection",
"multiprocessing.dummy",
"multiprocessing.managers",
"multiprocessing.pool",
"multiprocessing.sharedctypes",
"mutex",
"netrc",
"new",
"nis",
"nntplib",
"numbers",
"operator",
"optparse",
"os",
"os.path",
"ossaudiodev",
"parser",
"pdb",
"pickle",
"pickletools",
"pipes",
"pkgutil",
"platform",
"plistlib",
"popen2",
"poplib",
"posix",
"posixfile",
"pprint",
"profile",
"pstats",
"pty",
"pwd",
"py_compile",
"pyclbr",
"pydoc",
"quopri",
"random",
"re",
"readline",
"resource",
"rexec",
"rfc822",
"rlcompleter",
"robotparser",
"runpy",
"sched",
"select",
"sets",
"sgmllib",
"sha",
"shelve",
"shlex",
"shutil",
"signal",
"site",
"smtpd",
"smtplib",
"sndhdr",
"socket",
"spwd",
"sqlite3",
"ssl",
"stat",
"statvfs",
"string",
"stringprep",
"struct",
"subprocess",
"sunau",
"sunaudiodev",
"symbol",
"symtable",
"sys",
"sysconfig",
"syslog",
"tabnanny",
"tarfile",
"telnetlib",
"tempfile",
"termios",
"test",
"test.test_support",
"textwrap",
"thread",
"threading",
"time",
"timeit",
"token",
"tokenize",
"trace",
"traceback",
"ttk",
"tty",
"turtle",
"types",
"unicodedata",
"unittest",
"urllib",
"urllib2",
"urlparse",
"user",
"uu",
"uuid",
"videoreader",
"warnings",
"wave",
"weakref",
"webbrowser",
"whichdb",
"winsound",
"wsgiref",
"wsgiref.handlers",
"wsgiref.headers",
"wsgiref.simple_server",
"wsgiref.util",
"wsgiref.validate",
"xdrlib",
"xml",
"xml.dom",
"xml.dom.minidom",
"xml.dom.pulldom",
"xml.etree.ElementTree",
"xml.parsers.expat",
"xml.sax",
"xml.sax.handler",
"xml.sax.saxutils",
"xml.sax.xmlreader",
"xmlrpclib",
"zipfile",
"zipimport",
"zlib"]



# A part of the AegisBlade Python Client Library
# Copyright (C) 2019 Thornbury Organization, Bryan Thornbury
# This file may be used under the terms of the GNU Lesser General Public License, version 2.1.
# For more details see: https://www.gnu.org/licenses/lgpl-2.1.html


class Capability(object):
    """Enum-like class of capabilities available to be
    requested as part of an application.
    
    Capabilities are used to request functionality that cannot 
    be defined through the native package manager (pip in python's case).

    Attributes:
        xvfb: Requests that the application be run with xvfb-run.

        chrome: Requests that the application have the `google-chrome` and `chromedriver` 
            binaries available on the PATH. Also implies `Capability.xvfb`.

        firefox: Requests that the application have the `firefox` and `geckodriver` 
            binaries available on the PATH. Also implies `Capability.xvfb`.
    """

    xvfb = "xvfb"
    """Requests that the application be run with `xvfb-run`."""


    chrome = "browser-chrome"
    """Requests that the application have the `google-chrome` and `chromedriver` 
        binaries available on the PATH. Also implies `Capability.xvfb`"""

    firefox = "browser-firefox"
    """Requests that the application have the `firefox` and `geckodriver` 
            binaries available on the PATH. Also implies `Capability.xvfb`."""


# A part of the AegisBlade Python Client Library
# Copyright (C) 2019 Thornbury Organization, Bryan Thornbury
# This file may be used under the terms of the GNU Lesser General Public License, version 2.1.
# For more details see: https://www.gnu.org/licenses/lgpl-2.1.html

import base64
import time
import sys
import pickle

from typing import Any

from aegisblade.errors import ApplicationBuildError, JobError, \
    UnableToDeserializeReturnValueError, JobRuntimeError, JobNotFoundError
from aegisblade._internal.atrace import Trace
from aegisblade.ApiModels.responses import JobStatusResponse
from aegisblade._internal.api_consumer import _AegisBladeApiConsumer
from aegisblade.application import AegisBladeApplication


class AegisBladeJob(object):
    """An object for fetching job-related information from the API.
    
    This is typically returned by a call to `aegisblade.client.AegisBladeClient.run()` but may
    be constructed from a saved `AegisBladeJob.id` by calling the 
    `AegisBladeJob.create()` method.

    In AegisBlade, a job is conceptually a task to execute part
    of an application.

    Attributes:
        id (str): The GUID-like id of the job.

        application (aegisblade.application.AegisBladeApplication): A reference to the `aegisblade.application.AegisBladeApplication` object 
            that this job is running on. See `aegisblade.application.AegisBladeApplication` docs for 
            more details.

        job_type (str): The type of job represented by this object. Currently there
            is only one job type "InstantJob".
    """


    id = None                   # type: str
    """`str`: The GUID-like id of the job."""

    application = None          # type: AegisBladeApplication
    """`aegisblade.application.AegisBladeApplication`: A reference to the `aegisblade.application.AegisBladeApplication` object that this job is running on."""

    job_type = None             # type: str
    """`str`: The type of job represented by this object."""

    _api_consumer = None         # type: AegisBladeApiConsumer
    _poll_delay = 0.7            # type: int

    @classmethod
    def create(cls, client, job_id):
        # type: (AegisBladeClient, str) -> AegisBladeJob
        """Creates an AegisBladeJob instance from a job id.

        This method will contact the API to fetch information about the job
        then return an AegisBladeJob instance.

        Args:
            client (aegisblade.client.AegisBladeClient): The AegisBlade client. The default client is imported using 'from aegisblade import aegisblade'.
            job_id (str): The id of the job corresponding to the `AegisBladeJob.id` instance property.

        Raises:
            `aegisblade.errors.JobNotFoundError`: A job with the specified id was not found.
        """
        api_consumer = client._get_api_consumer()    # type: AegisBladeApiConsumer
        
        try:
            job_info = api_consumer.job_status(job_id)
        except Exception as e:
            if 'job not found' in str(e).lower():
                raise JobNotFoundError(job_id)

            raise

        return cls(client._get_api_consumer(), job_id, job_info.applicationId, job_info.jobType)

    def __init__(self, api_consumer, job_id, application_id, job_type):
        # type: (AegisBladeApiConsumer, str, str, str) -> None
        """The constructor is not recommended for direct use. 
        Use `AegisBladeJob.create()` to get an `AegisBladeJob` instance.
        """
        self._api_consumer = api_consumer
        self.id = job_id

        self.application = AegisBladeApplication(api_consumer, application_id)
        self.job_type = job_type

    def get_status(self):
        # type: () -> JobStatusResponse
        """
        Retrieves the status of this job from AegisBlade servers.

        Returns:
            `aegisblade.ApiModels.responses.JobStatusResponse`:  An object detailing the status of the job and any errors encountered
        """
        job_status_response = self._api_consumer.job_status(self.id)
        return job_status_response

    def get_logs(self):
        # type: () -> str
        """
        Retrieves the stdout & stderr logs output by the job while running.
        
        Returns:
            `str`: The stdout & stderr logs output by the job while running.
        """
        logs = self._api_consumer.job_logs(self.id)
        return logs

    def is_completed(self):
        # type: () -> bool
        """
        Returns whether or not the job has completed running. 
        
        Returns:
            `bool`: Whether or not the job has completed running.
        """
        status = self.get_status()

        return self._is_completed(status)

    def get_return_value(self, timeout=None):
        # type: (float) -> Any
        """Waits for the job to complete, and then fetches the serialized return value, deserializes it and returns it.

        Args:
            timeout (float): Timeout in seconds for waiting for job to complete, defaults to no timeout.

        Returns:
            `Any`: The deserialized value returned by the job's target Callable.

        Raises:
            `aegisblade.errors.JobRuntimeError`: The job raised an exception while running. The runtime exception is captured in `aegisblade.errors.JobRuntimeError.internal_error`.

            `aegisblade.errors.UnableToDeserializeReturnValueError`: The return value was unable to be deserialized.

        """
        self.wait_until_completed(timeout=timeout)

        serialized_return_value_str = self._api_consumer.job_return_value(self.id)

        try:
            return_value = pickle.loads(
                base64.b64decode(serialized_return_value_str))
        except:
            raise UnableToDeserializeReturnValueError(self.id)

        if issubclass(return_value.__class__, BaseException):
            Trace.debug("[DEBUG] Exception as return value.")
            Trace.debug(return_value)
            raise JobRuntimeError(self.id, return_value)

        return return_value

    def wait_until_completed(self, timeout=None):
        # type: (float) -> JobStatusResponse
        """Waits for the job to complete.

        Args:
            timeout (float): Timeout in seconds for waiting for job to complete, defaults to no timeout.
        """
        start_time = time.time()

        Trace.verbose("[INFO] Waiting until job completion ({0})".format(self.id))

        while True:
            status = self.get_status()

            if self._is_completed(status):
                Trace.verbose("[INFO] Job completed. ({0})".format(self.id))
                Trace.debug("[DEBUG] Job completed ({0}), status output: {1}".format(self.id, status))

                self._raise_job_error_if_applicable(status)

                return status

            time.sleep(self._poll_delay)

            if timeout is not None and (time.time() - start_time) > timeout:
                raise Exception("Timed out while waiting for job to complete")

    def _is_completed(self, status):
        # type: (JobStatusResponse) -> bool
        completed_states = ["finished", "canceled", "error"]

        return status.jobStatus.lower() in completed_states

    def _has_error(self, status):
        # type: (JobStatusResponse) -> bool
        error_states = ["error"]
        return status.jobStatus.lower() in error_states

    def _has_application_error(self, status):
        # type: (JobStatusResponse) -> bool
        if not self._has_error(status):
            return False

        return 'application failed to build' in status.errorDescription.lower()

    def _raise_job_error_if_applicable(self, status):
        if self._has_application_error(status) and self.application is not None:
            application_status = self.application.get_status()

            Trace.debug("[DEBUG] Error building application.")
            Trace.debug("[DEBUG] Application Status: {0}".format(application_status))

            raise ApplicationBuildError(self.application.id, application_status.errorDescription)

        if self._has_error(status):
            Trace.debug("[DEBUG] Error running job.")
            Trace.debug("[DEBUG] " + status.errorDescription)

            raise JobError(self.id, status.errorDescription)



# A part of the AegisBlade Python Client Library
# Copyright (C) 2019 Thornbury Organization, Bryan Thornbury
# This file may be used under the terms of the GNU Lesser General Public License, version 2.1.
# For more details see: https://www.gnu.org/licenses/lgpl-2.1.html

import os

from aegisblade._internal.environment import Environment


class Trace(object):
    """Internal class for outputting debug information to the console."""
    @classmethod
    def debug_output_level(cls):
        return Environment.debug_output is not None

    @classmethod
    def verbose_output_level(cls):
        return cls.debug_output_level() or Environment.verbose_output is not None

    @classmethod
    def debug(cls, message):
        if cls.debug_output_level():
            print(message)

    @classmethod
    def verbose(cls, message):
        if cls.verbose_output_level():
            print(message)



# A part of the AegisBlade Python Client Library
# Copyright (C) 2019 Thornbury Organization, Bryan Thornbury
# This file may be used under the terms of the GNU Lesser General Public License, version 2.1.
# For more details see: https://www.gnu.org/licenses/lgpl-2.1.html

import sys
import os
import imp


def get_std_lib_modules():
    current_version_str = '{0}.{1}'.format(*sys.version_info)
    return get_python_version_std_lib_modules(current_version_str)


def get_python_version_std_lib_modules(python_version):
    file_dir = os.path.dirname(os.path.abspath(__file__))
    target_file = os.path.join(file_dir, '_cached_stdlib_modules','v{0}.py'.format(python_version.replace('.', '_')))

    if not os.path.isfile(target_file):
        raise Exception(
            "Python version {0} not yet supported. "
            "See https://github.com/brthor/aegisblade for more info.".format(python_version))

    file_module = imp.load_source('stdlib', target_file)
    return file_module.__content__

    # std_lib_modules = []
    # with open(target_file, 'r') as f:
    #     for line in f:
    #         module_name = line.replace("\n", "").replace("\r", "")
    #         std_lib_modules.append(module_name)

    # return std_lib_modules



# A part of the AegisBlade Python Client Library
# Copyright (C) 2019 Thornbury Organization, Bryan Thornbury
# This file may be used under the terms of the GNU Lesser General Public License, version 2.1.
# For more details see: https://www.gnu.org/licenses/lgpl-2.1.html

import hashlib
import imp
import os
import re
import sys


from aegisblade._internal.atrace import Trace
from aegisblade._internal.application_context_file import ApplicationContextFile
from aegisblade._internal.application_execution_context import ApplicationExecutionContext
from aegisblade._internal.application_package_info import ApplicationPackageInfo
from aegisblade._internal.cached_std_lib_modules import get_std_lib_modules
from aegisblade._internal.command import Command
from aegisblade.__version__ import __version__


class _ApplicationExecutionContextCollector(object):
    """Internal class used to collect the files and data required to 
    build an executable version of this application.
    """

    def collect(self, libraries):
        # type: (dict) -> ApplicationExecutionContext

        pip_packages = self.get_pip_packages(libraries)

        application_execution_context = ApplicationExecutionContext(
            language="python",
            source_runtime_identifier=self.get_platform(),
            target_runtime_name="python",
            target_runtime_version="{0}.{1}".format(sys.version_info.major, sys.version_info.minor),
            packages=pip_packages,
            files=self.get_application_files(
                set([p.name for p in pip_packages]),
                self.get_std_lib_modules(),
                libraries
            ),
            package_manager_source_feeds=[],
            client_version=__version__
        )

        return application_execution_context

    def get_platform(self):
        try:
            import platform
            return platform.platform()
        except:
            return sys.platform

    def get_pip_packages(self, libraries):
        '''
        Get all installed packages (in the current environment)
        '''

        # Workaround for Ubuntu Bug
        # https://stackoverflow.com/questions/39577984/what-is-pkg-resources-0-0-0-in-output-of-pip-freeze-command
        excluded_packages = {'pkg-resources'}

        # Exclude also packages defined in 'libraries'
        for library_package, _ in libraries.items():
            excluded_packages.add(library_package)

        pip_process = Command("pip", ["freeze"]).environment_dict(os.environ).execute()

        pip_packages = []
        for line in pip_process.stdout.split("\n"):
            line = line.strip()
            if line == "":
                continue

            if line.startswith("-e"):
                pip_packages.append(ApplicationPackageInfo(line, None))
            else:
                name, version = line.split("==")

                if name not in excluded_packages:
                    pip_packages.append(ApplicationPackageInfo(name, version))

        return pip_packages

    def get_std_lib_modules(self):
        return get_std_lib_modules()

    def get_application_files(self, pip_package_names, std_library_modules, libraries):
        root_path = self.find_project_root()

        module_conditions = [
            lambda module: 'site-packages' not in module[1],  # Not a site package
            lambda module: not module[0].split('.')[0] in pip_package_names,
            # Not part of an installed package (backup), (optional)
            lambda module: not module[0].split('.')[0] in std_library_modules,
            # Not part of an any stdlib package, (optional)
            lambda module: not module[0] in std_library_modules,  # Not a stdlib package, (optional)
            lambda module: not module[0] in pip_package_names,  # Not a pip package, (optional),
            # lambda module: not module[1].startswith(os.environ.get("VIRTUAL_ENV", None)),
            # Not in the virtual env (some random files)
            lambda module: not '/pycharm.app/' in module[1].lower(),  # Not in pycharm (pycharm debugger),
            lambda module: not module[1].lower().startswith("/system"),   # Not in the OSX system directory
            lambda module: not module[1].lower().startswith("/usr/lib/python")  # Not in Linux Sys directory
        ]

        if os.environ.get("VIRTUAL_ENV", None):
            module_conditions.append(lambda module: not module[1].startswith(os.environ.get("VIRTUAL_ENV")))

        filtered_modules = [m for m in self.get_application_modules() if all(map(lambda condition: condition(m), module_conditions))]

        files = []
        for module in filtered_modules:
            name, path = module
            relative_path_to_root = os.path.relpath(path, root_path)

            if not self.is_subdir(path, root_path):
                found_library = None
                if libraries:
                    for _, library_path in libraries.items():
                        if path.startswith(library_path):
                            Trace.verbose("[INFO] Adding library module to aegisblade_lib {0} ({1})".format(path, library_path))
                            found_library = library_path
                            break

                if found_library:
                    lib_relative_path = os.path.relpath(path, found_library)
                    relative_path_to_root = os.path.join('aegisblade_lib', lib_relative_path)
                else:
                    Trace.verbose("WARNING. Skipping module outside project root directory. ({0})".format(path))
                    continue

            file_contents = None
            with open(path, 'rb') as f:
                file_contents = f.read()

            file_hash = hashlib.sha256()
            file_hash.update(file_contents)
            file_hash = file_hash.hexdigest()

            files.append(ApplicationContextFile(path, relative_path_to_root, file_hash, file_contents, len(file_contents)))

        return files

    def is_subdir(self, test_path, root_path):
        path = os.path.realpath(test_path)
        directory = os.path.realpath(root_path)

        relative = os.path.relpath(path, directory)
        return not (relative == os.pardir or relative.startswith(os.pardir + os.sep))

    def normalize_path(self, path_to_normalize):
        if path_to_normalize is None:
            return None

        normalized_path = os.path.normpath(
            os.path.normcase(
                os.path.abspath(
                    os.path.expanduser(
                        path_to_normalize
                    )
                )
            )
        )

        return normalized_path

    def get_filtered_sys_path(self):
        non_included_paths = []

        python_path = os.path.dirname(sys.executable)
        non_included_paths.append(self.normalize_path(python_path))

        virtual_env_path = os.environ.get("VIRTUAL_ENV", None)
        if virtual_env_path:
            non_included_paths.append(self.normalize_path(virtual_env_path))

        try:
            homebrew_prefix_cmd = Command("brew", ["--prefix"])\
                .environment_dict(os.environ)\
                .execute()

            if not homebrew_prefix_cmd.failed():
                homebrew_prefix = homebrew_prefix_cmd.stdout.replace("\n", "")
                homebrew_path = homebrew_prefix + "/Cellar"
                non_included_paths.append(self.normalize_path(homebrew_path))

        except OSError:
            # Expected on non-mac or non-brew systems
            pass

        path_components = filter(os.path.isdir,
                                map(self.normalize_path,
                                    sys.path[:]))

        for filtered_path in non_included_paths:
            path_components = filter(lambda p: not p.startswith(filtered_path), path_components)

        return path_components

    def get_py_file_suffix_or_none(self, file):
        for suffix in imp.get_suffixes():
            if file[-len(suffix[0]):] == suffix[0]:
                return suffix
        return None

    def get_application_modules(self):
        path = self.get_filtered_sys_path()
        # print("path: " + str(path))

        modules = {}

        for p in path:
            modules.update(self.get_modules_in_directory(p))

        items = list(modules.items())
        items.sort()
        return items

    def get_modules_in_directory(self, directory, d=1):
        # get modules in a given directory
        modules = {}
        for f in os.listdir(directory):
            f = os.path.join(directory, f)
            if os.path.isfile(f):
                m, e = os.path.splitext(f)
                suffix = self.get_py_file_suffix_or_none(f)
                if not suffix:
                    continue
                m = os.path.basename(m)
                if re.compile("(?i)[a-z_]\w*$").match(m):
                    # if suffix[2] == imp.C_EXTENSION:
                    #     # check that this extension can be imported
                    #     try:
                    #         __import__(m)
                    #     except ImportError:
                    #         continue
                    modules[m] = f
            elif os.path.isdir(f):
                m = os.path.basename(f)
                if os.path.isfile(os.path.join(f, "__init__.py")):
                    for mm, f in self.get_modules_in_directory(f, d=d+1).items():
                        modules[m + "." + mm] = f
        return modules

    def find_project_root(self):
        # TODO: Is there a better choice than CWD?
        return self.normalize_path(os.getcwd())



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

# A part of the AegisBlade Python Client Library
# Copyright (C) 2019 Thornbury Organization, Bryan Thornbury
# This file may be used under the terms of the GNU Lesser General Public License, version 2.1.
# For more details see: https://www.gnu.org/licenses/lgpl-2.1.html

from __future__ import absolute_import
from __future__ import print_function

import io
import re
from glob import glob
from os.path import basename
from os.path import dirname
from os.path import join
from os.path import splitext

import os

from setuptools import find_packages
from setuptools import setup


this_file_dir = os.path.dirname(
    os.path.abspath(
        os.path.realpath(__file__)))


def read(*names, **kwargs):
    return io.open(
        join(dirname(__file__), *names),
        encoding=kwargs.get('encoding', 'utf8')
    ).read()

def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__\s*=\s*['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


setup(
    name='aegisblade',
    version=find_version(this_file_dir, "aegisblade", "__version__.py"),
    license='LGPL2.1',
    description='Deploy & run your code in a single function call.',

    long_description=read(os.path.join(this_file_dir, "README.md")),
    long_description_content_type='text/markdown',

    author='AegisBlade',
    author_email='welovedevs@aegisblade.com',

    url='https://www.aegisblade.com/',
    project_urls={
        'Github': 'https://github.com/aegisblade/aegis-python',
        'Bug Reports': 'https://github.com/aegisblade/aegis-python/issues',
        'Docs': 'https://www.aegisblade.com/docs',
        'Reference': 'https://www.aegisblade.com/docs/reference/python'
    },

    install_requires=[
        'requests==2.20.1', 'typing==3.6.4', 'cloudpickle==1.2.2'
    ],

    python_requires=">=2.7",

    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,

    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)',
        'Operating System :: POSIX :: Linux',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet',
        'Topic :: System :: Distributed Computing',
        'Topic :: Software Development :: Libraries :: Python Modules'
    ],
    keywords='aegisblade server aws ec2 devops infrastructure as code IaC serverless kubernetes',
)


# A part of the AegisBlade Python Client Library
# Copyright (C) 2019 Thornbury Organization, Bryan Thornbury
# This file may be used under the terms of the GNU Lesser General Public License, version 2.1.
# For more details see: https://www.gnu.org/licenses/lgpl-2.1.html

import os

class _EnvironmentContainer(object):
    """Internal class wrapping access to environment variables."""

    @property
    def api_key(self):
        return os.environ.get("AEGISBLADE_API_KEY", None)

    @property
    def api_endpoint(self):
        return os.environ.get("AEGISBLADE_API_ENDPOINT", None)

    @property
    def default_hostdriver(self):
        return os.environ.get("AEGISBLADE_DEFAULT_HOSTDRIVER", None)

    @property
    def debug_output(self):
        return os.environ.get("AEGISBLADE_DEBUG_OUTPUT", None)
        
    @property
    def verbose_output(self):
        return os.environ.get("AEGISBLADE_VERBOSE_OUTPUT", None)

    @property
    def libraries(self):
        # type: () -> dict
        value = os.environ.get("AEGISBLADE_LIBRARIES", None)
        if value is None:
            return None

        library_dict = dict()

        for library_package_and_path in value.split(":"):
            try:
                library_package, library_path = library_package_and_path.split("=")
                library_dict[library_package] = library_path
            except Exception:
                print("Invalid AEGISBLADE_LIBRARIES format. Expected: \"package1=/some/path/1:package2=/some/path/2\"")
                raise
        
        return library_dict

    @property
    def verify_ssl(self):
        value = os.environ.get("AEGISBLADE_VERIFY_SSL", None)
        if value is None:
            return None

        if value.lower() == "true" or value == "1":
            return True
        
        return False

Environment = _EnvironmentContainer()